/*
 * $Id: statsurf.c,v 1.1 1999/04/12 16:36:44 altpeter Exp altpeter $
 * $Source: /usr/home/altpeter/statsurf/RCS/statsurf.c,v $
 *
 * $Log: statsurf.c,v $
 * Revision 1.1  1999/04/12 16:36:44  altpeter
 * Initial revision
 *
 */

/* StatSurf ******** Netsurf.DE user statistic service *******************\
*                                                                         *
* Copyright (c) 1999 Thomas 'Engerim' Kuiper and Frank 'Freddy' Altpeter  *
* Comments to: irc-develop@dachbu.de                                      *
* This program is distributed in the hope that it will be useful,         *
* but WITHOUT ANY WARRANTY; without even the implied warranty of          *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE and stuff.          *
*                                                                         *
* Some code from tkserv.c and other well known open source programs       *
* like fetchmail for the nice socketthingies.                             *
**************************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>
/* #include <tcpd.h> */
#include "servdef.h"
#include "config.h"
#include "statsurf.h"

#define SERVICE_VERSION "0.62"
#define SERVICE_MASK1 SERVICE_WANT_PREFIX + SERVICE_WANT_RQUIT + SERVICE_WANT_USER
#define SERVICE_MASK2 SERVICE_WANT_RQUIT + SERVICE_WANT_USER
#define AUTHORS "Frank Altpeter / Thomas Kuiper"

static int gl_users[1440]; /* 1440/60 = 24 hours,            */
static int tr_users[1440]; /* track every minute for the log */
static int sockserv;

static int total_users; /* total users online right now */
static int total_tr_users; /* total traced users online right now */

static long t_logins;
static long t_quits;
static long t_squery;
static long t_rcvd;
static long t_send;


int SockOpen(const char *host, int clientPort, const char *options)
{
    int sock;
    unsigned long inaddr;
    struct sockaddr_in ad;
    struct hostent *hp;

    memset(&ad, 0, sizeof(ad));
    ad.sin_family = AF_INET;

    /* we'll accept a quad address */
#ifndef INET_ATON
    inaddr = inet_addr(host);
    if (inaddr != 0)
        memcpy(&ad.sin_addr, &inaddr, sizeof(inaddr));
    else
#else
    if (!inet_aton(host, &ad.sin_addr))
#endif /* INET_ATON */
    {
        hp = gethostbyname(host);

	/*
	 * Add a check to make sure the address has a valid IPv4 or IPv6
	 * length.  This prevents buffer spamming by a broken DNS.
	 */
        if (hp == NULL || (hp->h_length != 4 && hp->h_length != 8))
            return -1;

	/*
	 * FIXME: make this work for multihomed hosts.
	 * We're toast if we get back multiple addresses and h_addrs[0]
	 * (aka h_addr) is not one we can actually connect to; this happens
	 * with multi-homed boxen.
	 */
        memcpy(&ad.sin_addr, hp->h_addr, hp->h_length);
    }
    ad.sin_port = htons(clientPort);
    
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0)
        return -1;
    if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) < 0)
    {
	close(sock);
        return -1;
    }

    return(sock);
}


int SockPrintf(int sock, char* format, ...)
{
    va_list ap;
    /* the sick thing here is we dont know how much the buffer should be */
    char buf[1024];

    va_start(ap, format) ;
#ifdef HAVE_VSNPRINTF
    vsnprintf(buf, sizeof(buf), format, ap);
#else
    vsprintf(buf, format, ap);
#endif
    va_end(ap);
    return SockWrite(sock, buf, strlen(buf));

}

int SockWrite(int sock, char *buf, int len)
{
    int n, wrlen = 0;

    while (len)
    {
        n = write(sock, buf, len);
        if (n <= 0)
            return -1;
        len -= n;
	wrlen += n;
	buf += n;
    }
    t_send = t_send + wrlen;
    return wrlen;
}

int SockRead(int sock, char *buf, int len)
{
    char *newline, *bp = buf;
    int n;

    if (--len < 1)
	return(-1);
    do {
	/* 
	 * The reason for these gymnastics is that we want two things:
	 * (1) to read \n-terminated lines,
	 * (2) to return the true length of data read, even if the
	 *     data coming in has embedded NULS.
	 */
	if ((n = recv(sock, bp, len, MSG_PEEK)) <= 0)
	    return(-1);
	if ((newline = memchr(bp, '\n', n)) != NULL)
	    n = newline - bp + 1;
	if ((n = read(sock, bp, n)) == -1)
	    return(-1);
	bp += n;
	len -= n;
    } while 
	    (!newline && len);
    *bp = '\0';
    t_rcvd = t_rcvd + len;
    return bp - buf;
}

int SockPeek(int sock)
/* peek at the next socket character without actually reading it */
{
    int n;
    char ch;

    if ((n = recv(sock, &ch, 1, MSG_PEEK)) == -1)
	return -1;
    else
	return(ch);
}

int main(int argc, char *argv[])
{
    FILE *fp;

    init_signals();
        
    connectservice();    

    printf("daemonized.\n");

    switch (fork())
        {
            case -1:
                perror("fork()");
                exit(3);
            case 0:
                close(STDIN_FILENO);
                close(STDOUT_FILENO);
                close(STDERR_FILENO);
	        if (setsid() == -1)
                    exit(4);
            break;
            default:
            return 0;
        }

    fp = fopen(PID_FILE,"w");
    if (fp != NULL)
    {
    fprintf (fp,"%i\n",getpid());
    fclose(fp);
    }
    alarm(60);
    communicate();
    

}

void init_signals()
{
    struct sigaction act;
    act.sa_handler = onhup;
    (void)sigemptyset(&act.sa_mask);
    (void)sigaddset(&act.sa_mask, SIGHUP);
    (void)sigaddset(&act.sa_mask, SIGALRM);
    (void)sigaddset(&act.sa_mask, SIGUSR1);
    (void)sigaddset(&act.sa_mask, SIGUSR2);
    (void)sigaction(SIGHUP, &act, NULL);
    (void)sigaction(SIGALRM, &act, NULL);
    (void)sigaction(SIGUSR1, &act, NULL);
    (void)sigaction(SIGUSR2, &act, NULL);
    (void)signal(SIGHUP, onhup);
    (void)signal(SIGALRM, onhup);
    (void)signal(SIGUSR1, onhup);
    (void)signal(SIGUSR2, onhup);
}



static void onhup(int s)
{
FILE *fp;
int i,tpos;
int th;
int tm;
time_t ttime;
struct tm *tmm;

if ((s == SIGHUP) || (s == SIGUSR1))
{
AddDebug("Got SIGHUP, reinitialing signals...\n");
    init_signals();
AddDebug("Done. Writing connlog...\n");
/* connlog gets written and cleared, gnuplot lauched */
fp = fopen (OUTFILE,"w");
if (fp != NULL)
    {
    for (i = 0;i<1440;i++)
	{
	th = i / 60;
	tm = th * 60 - i;
	tm = abs(tm);
	fprintf (fp,"%02i:%02i %i %i %i\n",th,tm,i,gl_users[i],tr_users[i]);
	if (s == SIGHUP)
	    {
    	    gl_users[i] = 0;
	    tr_users[i] = 0;
	    }
	}
    fclose(fp);
    }
system(STATSCRIPT);
communicate();
}

if (s == SIGALRM)
{
    AddDebug("Got SIGALRM, reinitialing signals...\n");
    init_signals();
    ttime = time(NULL);
    tmm = localtime(&ttime);
    tmm = tmm + STIME_OFFSET;
    tpos = tmm->tm_hour * 60+tmm->tm_min;
    gl_users[tpos] = total_users;
    tr_users[tpos] = total_tr_users;
    alarm(60);
    communicate();
}
if (s == SIGUSR2)
{
    AddDebug("Got SIGUSR2, reconnecting...\n");
    init_signals();
    connectservice();
}
return;
}

void communicate()
{
    int rcvd,i,tpos;
    char cbuf[400];
    char buf[400];
    char param1[400];
    char param2[400];
    char param3[400];
    char param4[400];
    char param5[400];
    time_t ttime;
    struct tm *tmm;
    rcvd = 0;
    while(rcvd != -1)
    {
    rcvd = SockRead(sockserv, cbuf, sizeof(cbuf));
    AddDebug(cbuf);
    sscanf (cbuf,"%s %s %s %s %s",param1,param2,param3,param4,param5);

    if (strcmp(param1,"PING") == 0)
    {
	SockPrintf(sockserv,"PONG %s\n",param2);
    }

    if (strcmp(param2,"USER") == 0)
    {
    
        total_users++;
	t_logins++;
	
	/* determine the current time and userhost for the log */
        ttime = time(NULL);
	tmm = localtime(&ttime);
	tmm = tmm + STIME_OFFSET;
        tpos = tmm->tm_hour * 60+tmm->tm_min;
        sprintf(cbuf,"User %s from %s signed on at %i(%02i:%02i)\n",
	param1,param4,tpos,tmm->tm_hour,tmm->tm_min);
	for (i =0;i<strlen(param4);i++) tolower(param5[i]);
	AddDebug(cbuf);
	gl_users[tpos] = total_users;
	if (strstr(param4,SERVICE_TRACK) != 0)
	{
	  total_tr_users++;
	  tr_users[tpos] = total_tr_users;
	  
	}
    }

    if (strcmp(param2,"QUIT") == 0)
    {
	/* userquit */
        total_users--;
	t_quits++;
        ttime = time(NULL);
	tmm = localtime(&ttime);
	tmm = tmm + STIME_OFFSET;
        tpos = tmm->tm_hour * 60+tmm->tm_min;
        sprintf(cbuf,"User %s signed off at %i(%02i:%02i)\n",
	param1,tpos,tmm->tm_hour,tmm->tm_min);
	for (i =0;i<strlen(param1);i++) tolower(param1[i]);
	AddDebug(cbuf);
        /* current time set has now this count of users: */
	gl_users[tpos] = total_users; 
	
	if (strstr(param1,SERVICE_TRACK) != 0)
	{
	  total_tr_users--;
	  tr_users[tpos] = total_tr_users;
	}
    }
        
    if (strcmp(param2,"SQUERY") == 0)
    {
	for (i = 0;i<strlen(param4);i++) param4[i] = toupper(param4[i]);
	for (i = 1;i<strlen(param1)+1;i++) param1[i-1] = param1[i];
	t_squery++;

	if (strcmp(param4,":ADMIN") == 0)
	{
	    SockPrintf(sockserv,"NOTICE %s :Administrative info about %s:\n",
	        param1,SERVICE_NAME);
	    SockPrintf(sockserv,"NOTICE %s :%s <%s>\n",
	        param1,ADMIN_NAME,ADMIN_EMAIL);
	
	}
	
	if (strcmp(param4,":HELP") == 0)
	{
	    SockPrintf(sockserv,"NOTICE %s :No help for this service.\n"
		    "NOTICE %s :Try 'admin' or 'info'\n", param1,
	    param1);
	}
	if (strcmp(param4,":INFO") == 0)
	{
	    SockPrintf(sockserv,"NOTICE %s :Information about %s:\n",
	        param1,SERVICE_NAME);
	    SockPrintf(sockserv,"NOTICE %s :Programmed by %s\n",
	        param1, AUTHORS);
	    SockPrintf(sockserv,"NOTICE %s :For more information mail to %s\n",
	        param1, ADMIN_EMAIL);
		
	
	}
	if (strcmp(param4,":VERSION") == 0)
	{
	    SockPrintf(sockserv,"NOTICE %s :%s version %s\n",
	        param1,SERVICE_NAME,SERVICE_VERSION);
	}
	if (strcmp(param4,":STATUS") == 0)
	{
	    SockPrintf(sockserv,"NOTICE %s :%s version %s status:\n",
	        param1,SERVICE_NAME,SERVICE_VERSION);
	    SockPrintf(sockserv,"NOTICE %s :Current local/tracked users: %i %i\n",
	        param1,total_users,total_tr_users);
	    SockPrintf(sockserv,"NOTICE %s :Total user connects/quits: %i %i\n",
	        param1,t_logins,t_quits);
	    SockPrintf(sockserv,"NOTICE %s :Total SQUERYs: %i\n",
	        param1,t_squery);
	    SockPrintf(sockserv,"NOTICE %s :Bytes Send/Received: %i %i\n",
	        param1,t_rcvd,t_send);

	}
    }
    }
    
}

void AddDebug(char *cbuf)
{
#ifdef SDEBUG
    
    FILE *debugf;
    debugf = fopen (DEBUG_FILE,"a");
	if (debugf != NULL)
	{
	fprintf(debugf,"%s",cbuf);
	fclose(debugf);
	}
    
#endif
}

void connectservice()
{
    /* for security reasons reset all fields */
    int i;
    t_logins = 0;
    t_quits = 0;
    t_squery = 0;
    t_rcvd = 0;
    t_send = 0;
    total_users = 0;
    total_tr_users = 0;
    for (i = 0;i<1440;i++)
    {
	gl_users[i] = 0;
	tr_users[i] = 0;
    }

    sockserv = SockOpen(SERVER_NAME,SERVER_PORT, NULL);

    printf("Starting Service...\n");
        
    if (sockserv == -1)
    {
    printf ("Error connecting to IRC Server\n");
    exit(1);
    }
    printf("Sending password...\n");
    
    SockPrintf(sockserv,"PASS %s\n", SERVICE_PASS);
    printf("Introducing...\n");
    SockPrintf(sockserv,"SERVICE %s %s %s %i 0 :%s\n",
    SERVICE_NAME,SERVICE_HOST,SERVICE_DIST,SERVICE_MASK2,SERVICE_DESC);
    SockPrintf(sockserv,"SERVSET %i %i\n",SERVICE_MASK1,SERVICE_MASK2);
    
}